/*
   This is a sample C program that shows how an application can read a
   congruent special value dataset. This program will read one dataset,
   "SnowDepth", that contains actual data. Value of some dataset elements
   should be replaced by special value.

   Then, this program will read a congurnet dataset, "SpecialValues", that
   contains special values. For simplicity, we will assume that both datasets
   are 5-by-7 array, and their datatypes are float and unsigned char,
   respectively. Also, we assume that there are two special values; 1 and 2.
   1 represents "WaterFilled" and 2 represents "IceCovered". 

   After reading both datasets, this program will print some human-readable
   output. If none of special values are set for an element, the actual value
   stored in the "SnowDepth" will be printed. If some special values are set, a
   special character will be printed instead. In this example, '/' is for the
   "WaterFilled" special value, '\' is for the "IceCovered" special value, and
   'X' is for the overlapped special value.

   For more information, please refer to the following sections from RFC:
   Special Values in HDF5 http://www.hdfgroup.org/pubs/rfcs/RFC_Special_Values_in_HDF5.pdf
      Section 4.3.3 "Parallel Special Value Dataset"
      Section 4.2.2 "Non-Mutually Exclusive Special Values"
 */

#define H5_USE_16_API
#include "hdf5.h"
#include <stdio.h>
#include <stdlib.h>

#define errorexit()  _errorexit(__LINE__)

static void _errorexit(int line)
{
	fprintf(stderr, "errorexit at %d\n", line);
	exit(1);
}

static void read_file(const char *fname)
{
	hid_t fileid, spaceid, dsetid;
	hsize_t dimsize[2] = { 5, 7 };
	float data[5][7];
	unsigned char masks[5][7];
	int i, j;

	/* open the HDF5 file using read-only mode */
	if ((fileid = H5Fopen(fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0) errorexit();

	/* create a dataspace for both the actual dataset and the congruent dataset */
	if ((spaceid = H5Screate_simple(2, dimsize, NULL)) < 0) errorexit();

	/* read a dataset */
	{
		/* open the HDF5 dataset */
		if ((dsetid = H5Dopen(fileid, "SnowDepth")) < 0) errorexit();
		/* read data from the HDF5 dataset, and store them in a variable called "data" */
		if (H5Dread(dsetid, H5T_NATIVE_FLOAT, spaceid, spaceid, H5P_DEFAULT, data) < 0) errorexit();
		/* close the dataset */
		if (H5Dclose(dsetid) < 0) errorexit();
	}

	/* read a congruent dataset */
	{
		/* open the congruent dataset */
		if ((dsetid = H5Dopen(fileid, "SpecialValues")) < 0) errorexit();
		/* read special values and store them in a variable called "masks" */
		if (H5Dread(dsetid, H5T_NATIVE_UCHAR, spaceid, spaceid, H5P_DEFAULT, masks) < 0) errorexit();
		/* close the dataset */
		if (H5Dclose(dsetid) < 0) errorexit();
	}

	/* close the dataspace */
	if (H5Sclose(spaceid) < 0) errorexit();
	/* close the file */
	if (H5Fclose(fileid) < 0) errorexit();

	/* combine the actual dataset and the congruent dataset to get a final result */
	/* this process could be done when the dataset is written, too */
	/* for each row */
	for (i = 0; i < 5; ++i) {
		/* for each column */
		for (j = 0; j < 7; ++j) {
			/* for each element */
			unsigned char mask;

			/* "mask" is the special value that corresponds to this data */
			mask = masks[i][j];

			/* if special value is not set for this element, just show the actual element */
			if (mask == 0)
				printf(" %4.1f ", data[i][j]);
			/* if special value is marked for this element, show a special character */
			else {
				char ch;
				/* if both special value '2' and special value '1' are marked, show 'X' */
				if (mask & 1 && mask & 2) ch = 'X';
				/* if only special value '2' is marked, show '\' */
				else if (mask & 2) ch = '\\';
				/* if only special value '1' is marked, show '/' */
				else if (mask & 1) ch = '/';
				/* if you want to merge the special value into the actual data, "data",
				 * you may want to do something like, data[i][j] = -1; */
				printf("  %c   ", ch);
			}
		}
		printf("\n");
	}
}

int main(int argc, char **argv)
{
	read_file("congruent.c.h5");
	return 0;
}

/* vi:set ts=4 sw=4 sts=4: */
